#include <ESP8266WiFi.h>
#include <WiFiClientSecure.h>

/*  Copyright (c) 2020 Jaroslaw Weglinski
    Code licensed under the MIT license.*/


#define DEBUG 1
const unsigned char LED_0F[] =
{
  // star marked characters have incorrect codes
  // 0    1     2     3     4     5     6     7     8     9     A     *b     C     d     E     *F     -   ' '
  0xC0, 0xF9, 0xA4, 0xB0, 0x99, 0x92, 0x82, 0xF8, 0x80, 0x90, 0x8C, 0xBF, 0xC6, 0xA1, 0x86, 0xFF, 0xbf, 0xFF
};

unsigned char LED[8];
const unsigned char SCLK = D2;
const unsigned char RCLK = D3;
const unsigned char DIO = D4;

const char* ssid     = "***********************INSERT SSID HERE****************";
const char* password = "***********************INSERT PASSWD HERE *************";


//https://media.githubusercontent.com/media/microsoft/Bing-COVID-19-Data/master/data/Bing-COVID19-Data.csv
const char* covidHost = "media.githubusercontent.com";
const char* covidUrl = "/media/microsoft/Bing-COVID-19-Data/master/data/Bing-COVID19-Data.csv";


//converts number to 1 of 8 code
byte numberTo1of8(byte in) {
  return 0x1 << (in & 0x7);
}

//blanks LED dispay 
void LED8_Blank(void) {
  shiftOut(DIO, SCLK, MSBFIRST, 0xFF);
  shiftOut(DIO, SCLK, MSBFIRST, 0x0);
  digitalWrite(RCLK, LOW);
  digitalWrite(RCLK, HIGH);
  return;
}

//prints global buffer LED[] on LED display by using "double" multiplexation - one segment at time
void LED8_Display (void)
{
  byte digit = 0;
  for (digit = 0; digit < 8 * 8; digit++)
  {
    shiftOut(DIO, SCLK, MSBFIRST, LED[digit & 0x7] | ~numberTo1of8(digit >> 3));
    shiftOut(DIO, SCLK, MSBFIRST, numberTo1of8(digit));
    digitalWrite(RCLK, LOW);
    digitalWrite(RCLK, HIGH);
  }
  //blank display after last digit to prevent uneven brightness
  LED8_Blank();
}

//writes unsigned integer as buffer of 7-segment LED digit codes
void printBuf7Seg(unsigned char *buf, unsigned char buflen, unsigned int val, boolean zeroFill = false) {
  unsigned char fillChar = 0x11;
  unsigned int tmp = val;
  unsigned char currPos = 0;

  if (buf == NULL)
    return;
  if (zeroFill)
    fillChar = 0x0;
  while (tmp > 0 && currPos < buflen) {
    buf[currPos] = LED_0F[tmp % 10];
    tmp /= 10;
    currPos++;
  }
  while (currPos < buflen) {
    buf[currPos++] = LED_0F[fillChar];
  }
  if (tmp > 0) {
    //on display overflow set all decimal points on
    for (currPos = 0; currPos < buflen; currPos++)
      buf[currPos] &= 0x7f;
  }
}


boolean getData(unsigned int* current, unsigned int* diff) {
  boolean readMore = true;
  String line;
  String prevLine;
  WiFiClientSecure client;
  client.setInsecure();
  if (!client.connect(covidHost, 443)) {
#if defined(DEBUG)
    Serial.println("HTTPS connection failed");
#endif
    return false;
  }
  client.print(String("GET ") + covidUrl + " HTTP/1.1\r\n" +
               "Host: " + covidHost + "\r\n" +
               "User-Agent: Arduino\r\n" +
               "Connection: close\r\n\r\n");
  while (client.connected()) {
    line = client.readStringUntil('\n');
    if (line == "\r") {
#if defined(DEBUG)
      Serial.println("headers received");
#endif
      break;
    }
  }

  //read csv header
  line = client.readStringUntil('\n');
  //process data lines
  while (readMore) {
    prevLine = line;
    line = client.readStringUntil('\n');
#if defined(DEBUG)
    Serial.println(line);
#endif
    if (line.startsWith("ID"))
      continue;
    if (line.endsWith("Worldwide,,\r"))
      continue;

    unsigned int pos = 0;
    unsigned int posTotal = 0;
    unsigned int posIncr = 0;
    unsigned int posEnd = 0;
    unsigned char i;
    for (i = 0; i < 8; i++) {
      pos = prevLine.indexOf(',', pos + 1);
      switch (i) {
        case 5:
          posTotal = pos;
          break;
        case 6:
          posIncr = pos;
          break;
        case 7:
          posEnd = pos;
          break;
      }
    }
#if defined(DEBUG)
    Serial.println(prevLine.substring(posTotal + 1, posIncr));
    Serial.println(prevLine.substring(posIncr + 1, posEnd));
#endif
    *current = prevLine.substring(posTotal + 1, posIncr).toInt();
    *diff = prevLine.substring(posIncr + 1, posEnd).toInt();
    readMore = false;
  }
  return true;
}

boolean setSteps(unsigned int* pcovidLevel, unsigned int* pcovidDiff, double * pcovidStep) {
  while (!getData(pcovidLevel, pcovidDiff)) {
#if defined(DEBUG)
    Serial.println("Data load error");
#endif
    delay(10000);
  }
  *pcovidLevel = *pcovidLevel - *pcovidDiff;
  *pcovidStep =  *pcovidDiff / (24 * 3600);
  return true;
}

boolean setSteps(unsigned int* pcovidLevel, unsigned int* pcovidDiff, unsigned int* pcovidStep, unsigned int* pcovidStepTime) {
  while (!getData(pcovidLevel, pcovidDiff)) {
#if defined(DEBUG)
    Serial.println("Data load error");
#endif
    delay(10000);
  }
  *pcovidLevel = *pcovidLevel - *pcovidDiff;
  //find step with significant counter increment
  for (*pcovidStepTime = 1; *pcovidStepTime <= 20; (*pcovidStepTime)++) {
    *pcovidStep = *pcovidDiff * *pcovidStepTime / (24 * 3600);
    if (*pcovidStep > 10)
      break;
  }
  if (*pcovidStepTime > 20 )
    *pcovidStepTime = 20;

#if defined(DEBUG)
  Serial.print("starting level: ");
  Serial.println(*pcovidLevel);
  Serial.print("difference: ");
  Serial.println(*pcovidDiff);
  Serial.print("step: ");
  Serial.println(*pcovidStep);
  Serial.print("step interval: ");
  Serial.println(*pcovidStepTime);
  Serial.print("end level: ");
  Serial.println(*pcovidLevel + *pcovidDiff);
#endif
  return true;
}

unsigned long tm;
unsigned long tmsync;
unsigned int covidDiff;
unsigned int covidLevel;
double covidCounter;
double covidStep;


void setup ()
{
  pinMode(SCLK, OUTPUT);
  pinMode(RCLK, OUTPUT);
  pinMode(DIO, OUTPUT);
  covidLevel = 0;
  covidDiff = 0 ;

  printBuf7Seg(LED, 8, covidLevel, false);
  LED8_Blank();

  Serial.begin(115200);
  delay(100);

#if defined(DEBUG)
  Serial.println();
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);
#endif

  WiFi.mode(WIFI_STA);
  //WiFi.disconnect();
  //delay(100);

  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
#if defined(DEBUG)
    Serial.print(".");
#endif
  }

#if defined(DEBUG)
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());
  Serial.print("Netmask: ");
  Serial.println(WiFi.subnetMask());
  Serial.print("Gateway: ");
  Serial.println(WiFi.gatewayIP());
#endif
  setSteps(&covidLevel, &covidDiff, &covidStep);
  covidCounter = (double)covidLevel;
  tm = millis();
  tmsync = tm;
}

void loop()
{
  unsigned long tmnew;
  printBuf7Seg(LED, 8, (unsigned int)(covidCounter + 0.5), false);
  LED8_Display();
  yield();
  tmnew = millis();
  if (tmnew - tm > 1000) {
    covidCounter += covidStep;
    tm = tmnew;
  }
  //get new data once a day
  if (tmnew - tmsync > 24 * 3600 * 1000) {
    setSteps(&covidLevel, &covidDiff, &covidStep);
    covidCounter = (double)covidLevel;
    tmsync = tmnew;
  }

}
